@@ -26,7 +26,7 @@
 #include <config.h>
 
 #include <bonobo-activation/bonobo-activation-private.h>
-#include <bonobo-activation/bonobo-activation-i18n.h>
+#include <glib/gi18n-lib.h>
 #include <bonobo-activation/bonobo-activation-init.h>
 #include <bonobo-activation/Bonobo_ActivationContext.h>
 
@@ -46,6 +46,13 @@
 #include <stdlib.h>
 #include <fcntl.h>
 
+static GMutex *thread_lock = NULL;
+static GCond  *thread_cond = NULL;
+static GSList *running_activations = NULL;
+
+#define RUNNING_LIST_LOCK()   if (thread_lock) g_mutex_lock (thread_lock);
+#define RUNNING_LIST_UNLOCK() if (thread_lock) g_mutex_unlock (thread_lock);
+
 /* Whacked from gnome-libs/libgnorba/orbitns.c */
 
 #define IORBUFSIZE 2048
@@ -108,12 +115,12 @@ exe_activate_info_to_retval (EXEActivateInfo *ai, CORBA_Environment *ev)
         return retval;
 }
 
-static CORBA_Object
-scan_list (GSList *l, EXEActivateInfo *seek_ai, CORBA_Environment *ev)
+static EXEActivateInfo *
+find_on_list (EXEActivateInfo *seek_ai, CORBA_Environment *ev)
 {
-        CORBA_Object retval = CORBA_OBJECT_NIL;
+        GSList *l;
 
-        for (; l; l = l->next) {
+        for (l = running_activations; l; l = l->next) {
                 EXEActivateInfo *ai = l->data;
 
                 if (strcmp (seek_ai->exename, ai->exename))
@@ -127,29 +134,82 @@ scan_list (GSList *l, EXEActivateInfo *seek_ai, CORBA_Environment *ev)
                 } else if (seek_ai->environment || ai->environment)
                         continue;
 
+                return ai;
+        }
+
+        return NULL;
+}
+
+static CORBA_Object
+scan_list (EXEActivateInfo *seek_ai, CORBA_Environment *ev)
+{
+        EXEActivateInfo *ai;
+        CORBA_Object retval = CORBA_OBJECT_NIL;
+
+
+        if (thread_cond) {
+                /*
+                 * someone might have registered while we
+                 * dropped the main server lock => re-check,
+                 * inelegant but ...
+                 */
+#ifdef BONOBO_ACTIVATION_DEBUG
+                g_message ("Pre-check for ... '%s' \n", seek_ai->act_iid);
+#endif
+                retval = seek_ai->re_check (seek_ai->environment,
+                                            seek_ai->act_iid,
+                                            seek_ai->user_data);
+                if (retval != CORBA_OBJECT_NIL)
+                        return retval;
+        }
+
+        if (!(ai = find_on_list (seek_ai, ev)))
+                return CORBA_OBJECT_NIL;
+
+        if (thread_cond) {
+                while (ai && !ai->done) {
+#ifdef BONOBO_ACTIVATION_DEBUG
+                        g_message ("thread %p: activation of '%s' already pending, waiting ...\n",
+                                   g_thread_self(), seek_ai->act_iid);
+#endif
+                        /* Wait for something to happen */
+                        g_cond_wait (thread_cond, thread_lock);
+#ifdef BONOBO_ACTIVATION_DEBUG
+                        g_message ("thread %p: activation of '%s' woken up to retry ...\n",
+                                   g_thread_self(), seek_ai->act_iid);
+#endif
+                        ai = find_on_list (seek_ai, ev);
+                }
+        } else {
                 /* We run the loop too ... */
                 while (!ai->done)
                         g_main_context_iteration (NULL, TRUE);
-
-                if (!strcmp (seek_ai->act_iid, ai->act_iid)) {
-#ifdef BONOBO_ACTIVATION_DEBUG
-                        g_message ("Hit the jackpot '%s' '%s'",
-                                   seek_ai->act_iid, ai->act_iid);
-#endif
-                        retval = exe_activate_info_to_retval (ai, ev);
-                } else if (seek_ai->re_check) {
-                        /* It might have just registered the IID */
-#ifdef BONOBO_ACTIVATION_DEBUG
-                        g_message ("Re-check the thing ... '%s' '%s'",
-                                   seek_ai->act_iid, ai->act_iid);
-#endif
-                        retval = seek_ai->re_check (
-					seek_ai->environment,
-					seek_ai->act_iid,
-					seek_ai->user_data, ev);
-                }
         }
 
+        if (ai && !strcmp (seek_ai->act_iid, ai->act_iid)) {
+#ifdef BONOBO_ACTIVATION_DEBUG
+                g_message ("thread %p hit the jackpot '%s' '%s'\n",
+                           g_thread_self(), seek_ai->act_iid, ai->act_iid);
+#endif
+                retval = exe_activate_info_to_retval (ai, ev);
+                if (ev->_major != CORBA_NO_EXCEPTION) g_message ("URGH ! 3\n");
+        } else if (seek_ai->re_check) {
+                /* It might have just registered the IID */
+#ifdef BONOBO_ACTIVATION_DEBUG
+                g_message ("thread %p re-check for ... '%s' \n",
+                           g_thread_self(), seek_ai->act_iid);
+#endif
+                retval = seek_ai->re_check (seek_ai->environment,
+                                            seek_ai->act_iid,
+                                            seek_ai->user_data);
+        } else {
+#ifdef BONOBO_ACTIVATION_DEBUG
+                g_warning ("thread %p: very unusual dual activation failure: '%s'\n",
+                           g_thread_self(), seek_ai->act_iid);
+#endif
+        }
+                
+
         return retval;
 }
 
@@ -177,6 +237,9 @@ handle_exepipe (GIOChannel * source,
                                     _("Failed to read from child process: %s\n"),
                                     error->message);
                         g_error_free (error);
+#ifdef BONOBO_ACTIVATION_DEBUG
+                        fprintf (stderr, "b-a-f-s failed to read from child '%s'\n", error->message);
+#endif
                         error = NULL;
                         retval = FALSE;
                 } else if (status == G_IO_STATUS_EOF) {
@@ -188,9 +251,9 @@ handle_exepipe (GIOChannel * source,
                         retval = TRUE;
                 }
                 g_string_free (str, TRUE);
-        }
-        else
+        } else {                
                 retval = FALSE;
+        }
 
 	if (retval && !strncmp (data->iorbuf, "IOR:", 4))
 		retval = FALSE;
@@ -206,6 +269,32 @@ handle_exepipe (GIOChannel * source,
 	return retval;
 }
 
+void
+bonobo_activation_server_fork_init (gboolean threaded)
+{
+        if (threaded) {
+                thread_lock = g_mutex_new ();
+                thread_cond = g_cond_new ();
+        }
+}
+
+#ifndef G_OS_WIN32
+
+static void
+child_setup (gpointer user_data)
+{
+        int pipe_fd = GPOINTER_TO_INT (user_data);
+
+        /* unset close on exec */
+        fcntl (pipe_fd, F_SETFD, 0);
+}
+
+#else
+
+#define child_setup NULL
+
+#endif
+
 CORBA_Object
 bonobo_activation_server_by_forking (
 	const char                         **cmd_const,
@@ -219,10 +308,8 @@ bonobo_activation_server_by_forking (
 	gpointer                             user_data,
 	CORBA_Environment                   *ev)
 {
-        int i;
 	gint iopipes[2];
 	CORBA_Object retval = CORBA_OBJECT_NIL;
-        static GSList *running_activations = NULL;
         EXEActivateInfo ai;
         GError *error = NULL;
         GSource *source;
@@ -243,10 +330,16 @@ bonobo_activation_server_by_forking (
 #ifdef BONOBO_ACTIVATION_DEBUG
         ai.do_srv_output = getenv ("BONOBO_ACTIVATION_DEBUG_EXERUN");
 #endif
-                
+
+        RUNNING_LIST_LOCK();
         if (!use_new_loop &&
-            (retval = scan_list (running_activations, &ai, ev)) != CORBA_OBJECT_NIL)
+            (retval = scan_list (&ai, ev)) != CORBA_OBJECT_NIL) {
+                RUNNING_LIST_UNLOCK();
                 return retval;
+        }
+
+        if (thread_lock) /* don't allow re-enterancy in this thread */
+                use_new_loop = TRUE;
         
      	pipe (iopipes);
 
@@ -311,8 +404,8 @@ bonobo_activation_server_by_forking (
                 newenv [n] = NULL;
         }
 
-        cmd = g_strdupv (cmd_const);
         /* Pass the IOR pipe's write end to the child */ 
+        cmd = g_strdupv ((char **)cmd_const);
         if (fd_arg != 0)
         {
                 g_free (cmd[fd_arg]);
@@ -323,13 +416,18 @@ bonobo_activation_server_by_forking (
         ai.done = FALSE;
         
         running_activations = g_slist_prepend (running_activations, &ai);
+        RUNNING_LIST_UNLOCK();
 
 	/* Spawn */
         if (!g_spawn_async (NULL, (gchar **) cmd, newenv,
+#ifdef G_OS_WIN32
+                            /* win32 g_spawn doesn't handle child_setup, so
+                               leave all fds open */
                             G_SPAWN_LEAVE_DESCRIPTORS_OPEN |
+#endif
                             G_SPAWN_SEARCH_PATH |
                             G_SPAWN_CHILD_INHERITS_STDIN,
-                            NULL, NULL,
+                            child_setup, GINT_TO_POINTER (iopipes[1]),
                             NULL,
                             &error)) {
                 Bonobo_GeneralError *errval;
@@ -337,6 +435,9 @@ bonobo_activation_server_by_forking (
                                                     ":",
                                                     error->message,
                                                     NULL);
+#ifdef BONOBO_ACTIVATION_DEBUG
+                fprintf (stderr, "g_spawn_async error '%s'\n", error->message);
+#endif
                 g_error_free (error);
                 error = NULL;
 
@@ -347,32 +448,35 @@ bonobo_activation_server_by_forking (
 		CORBA_exception_set (ev, CORBA_USER_EXCEPTION,
 				     ex_Bonobo_GeneralError, errval);
 
+                RUNNING_LIST_LOCK();
                 running_activations = g_slist_remove (running_activations, &ai);
+                RUNNING_LIST_UNLOCK();
+
+                if (thread_cond)
+                        g_cond_broadcast (thread_cond);
+
+                close (iopipes[1]);
 
                 g_source_destroy (source);
                 g_source_unref (source);
                 
+                g_io_channel_shutdown (ai.ioc, FALSE, NULL);
                 g_io_channel_unref (ai.ioc);
                 
                 if (use_new_loop)
                         g_main_context_unref (context);
 
+                g_strfreev (newenv);
                 g_strfreev (cmd);
 
+                close (iopipes[0]);
+
 		return CORBA_OBJECT_NIL;
 	}
 
         close (iopipes[1]);
 
-        if (newenv) {
-                char **ep = newenv;
-
-                while (*ep) {
-                        g_free (*ep);
-                        ep++;
-                }
-                g_free (newenv);
-        }
+        g_strfreev (newenv);
         g_strfreev (cmd);
 
         /* Get the IOR from the pipe */
@@ -380,6 +484,13 @@ bonobo_activation_server_by_forking (
                 g_main_context_iteration (context, TRUE);
         }
 
+#ifdef BONOBO_ACTIVATION_DEBUG
+        g_message ("thread %p: broadcast activation of '%s' complete ...\n",
+                   g_thread_self(), act_iid);
+#endif
+        if (thread_cond)
+                g_cond_broadcast (thread_cond);
+
         g_source_destroy (source);
         g_source_unref (source);
         
@@ -389,7 +500,9 @@ bonobo_activation_server_by_forking (
         if (use_new_loop)
                 g_main_context_unref (context);
         
+        RUNNING_LIST_LOCK();
         running_activations = g_slist_remove (running_activations, &ai);
+        RUNNING_LIST_UNLOCK();
 
         retval = exe_activate_info_to_retval (&ai, ev); 
 
